#include <windows.h>
#include <string>
#include <vector>
#include <set>
#include <iostream>
#include <algorithm>
#include "boost/algorithm/string/split.hpp"
#include "boost/algorithm/string/join.hpp"
#include "boost/algorithm/string/classification.hpp"
#include "junction.h"

typedef struct juncpair
{
	std::string from;
	std::string to;
	int diff;
} juncpair;

bool diffjl(const juncpair & a, const juncpair & b)
{
	return a.diff > b.diff;
}

bool shortpath(std::string path, std::string & outpath)
{
	char tmppath[MAX_PATH];
	if (GetShortPathNameA(path.c_str(), tmppath, sizeof(tmppath)))
	{
		std::string smlpath(tmppath);
		outpath.swap(smlpath);
		return true;
	}
	return false;
}

std::string collapsepath(std::string inpath, std::vector<juncpair> & junctions) 
{
	std::vector<std::string> paths;
	boost::algorithm::split(paths, inpath, boost::algorithm::is_any_of(";"));

	std::set<std::string> discovered;
	std::vector<std::string> outpaths;
	for (std::vector<std::string>::const_iterator vit = paths.begin(); vit != paths.end(); ++vit)
	{
		std::string outpath;
		if (shortpath(*vit, outpath))
		{
			if (discovered.insert(outpath).second) 
			{
				outpaths.push_back(outpath);
			}
		}
		else
		{
			std::cerr << "[invalid entry '" << *vit << "']\n";
		}
	}

	// Now walk THESE, and see if any can be shrunk further with junctions
	if (!junctions.empty())
	{
		for (std::vector<std::string>::iterator pit = outpaths.begin(); pit != outpaths.end(); ++pit)
		{
			std::string & path1 = *pit;
			std::string p1l = path1;
			std::transform(p1l.begin(), p1l.end(), p1l.begin(), tolower);
			for (std::vector<juncpair>::iterator jit = junctions.begin(); jit != junctions.end(); ++jit)
			{
				std::string & jtgt = jit->to;
				std::string & jsrc = jit->from;
				if (p1l.length() >= jtgt.length() && jtgt == p1l.substr(0, jtgt.length())) 
				{
					path1 = jsrc + path1.substr(jtgt.length());
					break;
				}
			}
		}
	}

	return boost::algorithm::join(outpaths, ";");
}


int main(int argc, char* argv[])
{
	if (argc < 2) 
	{
		std::cerr << argv[0] << " fullpath [-j junction]+\n";
		return 1;
	}
	std::string path(argv[1]);
	std::vector<juncpair> junctions;
	int pc = 2;
	while (pc < argc-1)
	{
		// stricmp, dammit!
		if (std::string(argv[pc]) != "-j")
		{
			std::cerr << "Unknown flag [" << argv[pc] << "]\n";
			return 1;
		}
		std::string juncsrc = argv[pc+1];
		try
		{
			std::string junctgt = GetReparseTarget(juncsrc);
			juncpair jp;
			if (shortpath(juncsrc, jp.from) && 
				shortpath(junctgt, jp.to))
			{
				jp.diff = junctgt.length() - juncsrc.length();
				std::transform(jp.to.begin(), jp.to.end(), jp.to.begin(), tolower);
				if (jp.diff > 0) junctions.push_back(jp);
			}
			else
			{
				std::cerr << "[can't use junction '" << juncsrc << "']\n";
			}
		}
		catch (...)
		{
			// Ignore any we can't deal with.
			std::cerr << "[can't process junction '" << juncsrc << "']\n";
		}
		pc+=2;
	}
	// sort junctions by descending length of difference between src -> dest
	std::sort(junctions.begin(), junctions.end(), diffjl);
	std::string newpath = collapsepath(path,junctions);

	std::cerr << "Original path [" << path.length() <<"]:\n" << path << "\n\nNew path [" << newpath.length() << "]:\n";
	std::cout << newpath << "\n";
}
